// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  Email:
//    mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>
#include "../../util/EnterDataUtil.h"
#include "../../util/tiempo.h"

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

class MyDataStream : public IDataStream
{
public:
	MyDataStream(string inputFile) : m_pNext(0)
	{
		m_fin.open(inputFile.c_str());

		if (! m_fin)
			throw Tools::IllegalArgumentException("Input file not found.");

		id_type id;
		double lon, lat;
		double plow[2], phigh[2];

		m_fin >> id >> lon >> lat;

		if (m_fin.good())
		{
			plow[0] = lon; plow[1] = lat;
			phigh[0] = lon; phigh[1] = lat;
			Region r = Region(plow, phigh, 2);
			m_pNext = new RTree::Data(0, 0, r, id);
		}
	}

	virtual ~MyDataStream()
	{
		if (m_pNext != 0) delete m_pNext;
	}

	virtual IData* getNext()
	{
		if (m_pNext == 0) return 0;

		RTree::Data* ret = m_pNext;
		m_pNext = 0;

		id_type id;
		double lon, lat;
		double plow[2], phigh[2];

		m_fin >> id >> lon >> lat;

		if (m_fin.good())
		{
			plow[0] = lon; plow[1] = lat;
			phigh[0] = lon; phigh[1] = lat;
			Region r = Region(plow, phigh, 2);
			m_pNext = new RTree::Data(0, 0, r, id);
		}

		return ret;
	}

	virtual bool hasNext() throw (Tools::NotSupportedException)
	{
		return (m_pNext != 0);
	}

	virtual size_t size() throw (Tools::NotSupportedException)
	{
		throw Tools::NotSupportedException("Operation not supported.");
	}

	virtual void rewind() throw (Tools::NotSupportedException)
	{
		id_type id;
		double lon, lat;
		double plow[2], phigh[2];

		if (m_pNext != 0)
		{
			delete m_pNext;
			m_pNext = 0;
		}

		m_fin.seekg(0, std::ios::beg);

		m_fin >> id >> lon >> lat;

		if (m_fin.good())
		{
			plow[0] = lon; plow[1] = lat;
			phigh[0] = lon; phigh[1] = lat;
			Region r = Region(plow, phigh, 2);
			m_pNext = new RTree::Data(0, 0, r, id);
		}
	}

	std::ifstream m_fin;
	RTree::Data* m_pNext;
};

// example of a Visitor pattern.
// see RTreeQuery for a more elaborate example.
class MyVisitor : public IVisitor
{
private:
	long result[MAX_RESULTS];
	uint resultSize;
public:
	void visitNode(const INode& n) {}

	void visitData(const IData& d)
	{
		result[resultSize++] = d.getIdentifier();
		//cout << d.getIdentifier() << "; ";
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}

	void visitData(std::vector<const IData*>& v) {}
	
	void clearResult() {
		resultSize = 0;
	}
	void printResult() {
		printf("%ld elements: ", resultSize);
		uint j;
		for(j = 0; j < resultSize; j++) {
			printf("%ld; ", result[j]);
		}
		
		printf("\n");
	}
};

int main(int argc, char** argv)
{
	try
	{
		if (argc < 8)
		{
			cerr << "Usage: " << argv[0] << " input_file file_type [IdLonLat | ...] capacity query_type [intersection | 10NN | selfjoin] output_file query_input_file1 [query_input_filek] query_file_type [IdX1Y1X2Y2 | ...]." << endl;
			return -1;
		}

		uint32_t queryType = 0;
		uint32_t fileType = 0;
		uint32_t queryFileType = 0;

		if (strcmp(argv[4], "intersection") == 0) queryType = 0;
		else if (strcmp(argv[4], "10NN") == 0) queryType = 1;
		else if (strcmp(argv[4], "selfjoin") == 0) queryType = 2;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		if (strcmp(argv[2], "IdLonLat") == 0) fileType = 0;
		else
		{
			cerr << "Unknown file type." << endl;
			return -1;
		}

		if (strcmp(argv[argc - 1], "IdX1Y1X2Y2") == 0) queryFileType = 0;
		else
		{
			cerr << "Unknown query file type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open data file " << argv[1] << "." << endl;
			return -1;
		}

		cerr << "Creando RTree..." << endl;
		
		// Create a new storage manager
		IStorageManager* memorymanager = StorageManager::createNewMemoryStorageManager();

		//Create the data stream
		MyDataStream stream(argv[1]);

		// Create a new, empty, RTree with dimensionality 2, minimum load 70%, using "memorymanager" as
		// the StorageManager and the RSTAR splitting policy.
		id_type indexIdentifier;
		//ISpatialIndex* tree = RTree::createNewRTree(*memorymanager, 0.7, atoi(argv[3]), atoi(argv[3]), 2, SpatialIndex::RTree::RV_RSTAR, indexIdentifier);
		ISpatialIndex* tree = RTree::createAndBulkLoadNewRTree(
			RTree::BLM_STR, stream, 
			*memorymanager, 0.7, atoi(argv[3]), atoi(argv[3]), 2, SpatialIndex::RTree::RV_RSTAR, indexIdentifier);

/*		
		//read input elements
		uint number_of_input_elements;
		INPUT_POINT *input_data;
		if(fileType == 0) {
			readIdLonLat(argv[1], &input_data, &number_of_input_elements);
		}
		
		//insert elements into the RTree
		int i;
		double plow[2], phigh[2];
		for(i = 0; i < number_of_input_elements; i++) {
			plow[0] = input_data[i].point.lon; plow[1] = input_data[i].point.lat;
			phigh[0] = input_data[i].point.lon; phigh[1] = input_data[i].point.lat;
			Region r = Region(plow, phigh, 2);
			tree->insertData(0, 0, r, input_data[i].id);

			//if ((i % 1000) == 0)
			//	cerr << i << endl;
		}
		free(input_data);
*/ 
		cerr << "RTree Creado" << endl;
		
		//cerr << "Operations: " << i << endl;
		cerr << *tree;
		//cerr << "Index ID: " << indexIdentifier << endl;

		//bool ret = tree->isIndexValid();
		//if (ret == false) cerr << "ERROR: Structure is invalid!" << endl;
		//else cerr << "The stucture seems O.K." << endl;
		
		FILE* outputFile = fopen(argv[5], "a");
		uint k;
		for(k = 6; k < argc - 1; k++) {
			//read input queries
			uint number_of_input_queries;
			INPUT_MBR *input_queries;
			if(fileType == 0) {
				readIdX1Y1X2Y2(argv[k], &input_queries, &number_of_input_queries);
			}
			
			//execute queries
			double x1, x2, y1, y2;
		
			double time = 0;
		
			int i;
			double plow[2], phigh[2];
			for(i = 0; i < number_of_input_queries; i++) {
				plow[0] = input_queries[i].mbr.point1.lon; plow[1] = input_queries[i].mbr.point1.lat;
				phigh[0] = input_queries[i].mbr.point2.lon; phigh[1] = input_queries[i].mbr.point2.lat;
				Region r = Region(plow, phigh, 2);
				
				MyVisitor vis;

				tTiempo mitiempo1;
				tTiempo mitiempo2;
				
				register uint j;

				getTiempo(&mitiempo1);

				for(j = 0; j < 10; j++) {

				vis.clearResult();

/*			
				if (queryType == 0)
				{
*/			
					Region r = Region(plow, phigh, 2);
					tree->intersectsWithQuery(r, vis);
						// this will find all data that intersect with the query range.
/*
						}
				else if (queryType == 1)
				{
					Point p = Point(plow, 2);
					tree->nearestNeighborQuery(10, p, vis);
						// this will find the 10 nearest neighbors.
				}
				else
				{
					Region r = Region(plow, phigh, 2);
					tree->selfJoinQuery(r, vis);
				}
*/
				}
				
				getTiempo(&mitiempo2);
				time += ((mitiempo2.user - mitiempo1.user) / 10);
//			muestraTiempo(mitiempo1, mitiempo2, "tiempo");

				//cout << "Query " << input_queries[i].id << " solution: ";
				//vis.printResult();
			}
			fprintf(outputFile, "%lf;", time / number_of_input_queries);
			free(input_queries);		
		}

		fprintf(outputFile, "\n");
		fclose(outputFile);
		
		delete tree;
		delete memorymanager;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail trying to write the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
