// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  Email:
//    mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <algorithm>
#include <vector>

#include <SpatialIndex.h>
#include "../../util/EnterDataUtil.h"
#ifndef TIEMPO_H
#include "../../util/tiempo.h"
#endif
#include "SRTree.h"

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

#define INPUT_TYPE_ID_MBR 0
#define INPUT_TYPE_MBR 1

#define MAX_RESULTS 2000000

class MyDataStream : public IDataStream
{
public:
	MyDataStream(string inputFile, uint fileType) : globalID(0),m_pNext(0)
	{
		processed = 0;
		inputFileType = fileType;
		m_fin.open(inputFile.c_str());

		if (! m_fin)
			throw Tools::IllegalArgumentException("Input file not found.");

		id_type id;
		double low[2], high[2];

		if(inputFileType == INPUT_TYPE_ID_MBR) {
			m_fin >> id >> low[0] >> low[1] >> high[0] >> high[1];
		} else if(inputFileType == INPUT_TYPE_MBR) {
			m_fin >> low[0] >> low[1] >> high[0] >> high[1];
			id = globalID++;
		}

		if (m_fin.good())
		{
			Region r = Region(low, high, 2);
			m_pNext = new RTree::Data(0, 0, r, id);
		}
	}

	virtual ~MyDataStream()
	{
		if (m_pNext != 0) delete m_pNext;
	}

	virtual IData* getNext()
	{
		if (m_pNext == 0) return 0;

		RTree::Data* ret = m_pNext;
		m_pNext = 0;

		id_type id;
		double low[2], high[2];

		if(inputFileType == INPUT_TYPE_ID_MBR) {
			m_fin >> id >> low[0] >> low[1] >> high[0] >> high[1];
		} else if(inputFileType == INPUT_TYPE_MBR) {
			m_fin >> low[0] >> low[1] >> high[0] >> high[1];
			id = globalID++;
		}

		if (m_fin.good())
		{
			Region r = Region(low, high, 2);
			m_pNext = new RTree::Data(0, 0, r, id);
		}
		processed++;

		return ret;
	}

	virtual bool hasNext() throw (Tools::NotSupportedException)
	{
		return (m_pNext != 0);
	}

	virtual size_t size() throw (Tools::NotSupportedException)
	{
		return processed;
	}

	virtual void rewind() throw (Tools::NotSupportedException)
	{
		id_type id;
		double low[2], high[2];

		if (m_pNext != 0)
		{
			delete m_pNext;
			m_pNext = 0;
		}

		m_fin.seekg(0, std::ios::beg);

		if(inputFileType == INPUT_TYPE_ID_MBR) {
			m_fin >> id >> low[0] >> low[1] >> high[0] >> high[1];
		} else if(inputFileType == INPUT_TYPE_MBR) {
			m_fin >> low[0] >> low[1] >> high[0] >> high[1];
			id = globalID++;
		}

		if (m_fin.good())
		{
			Region r = Region(low, high, 2);
			m_pNext = new RTree::Data(0, 0, r, id);
		}
	}

	std::ifstream m_fin;
	RTree::Data* m_pNext;
	id_type globalID;
	size_t processed;
	uint inputFileType;
};

// example of a Visitor pattern.
// see RTreeQuery for a more elaborate example.
class MyVisitor : public IVisitor
{
private:
	long result[MAX_RESULTS];
	uint resultSize;
public:
	void visitNode(const INode& n) {}

	void visitData(const IData& d)
	{
		result[resultSize++] = d.getIdentifier();
		//cout << d.getIdentifier() << "; ";
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}

	void visitData(std::vector<const IData*>& v) {}
	
	void clearResult() {
		resultSize = 0;
	}
	void printResult() {
		qsort(result, resultSize, sizeof(long), cmplong);
		//printf("%ld", resultSize);
		printf("%ld elements: ", resultSize);
		
		uint j;
		for(j = 0; j < resultSize; j++) {
			printf("%ld; ", result[j]);
		}
		
		printf("\n");
	}
};

int main(int argc, char** argv)
{
	try
	{
		if (argc < 11)
		{
			cerr << "Usage: " << argv[0] << " input_file file_type [IdX1Y1X2Y2 | X1Y1X2Y2 | ...] capacity query_type [intersection | 10NN | selfjoin] output_file query_input_file1 [query_input_filek] query_file_type [IdX1Y1X2Y2 | ...] index_type [REG|COMP|HYB] extra_param disc_factor test|time." << endl;
			return -1;
		}

		uint32_t queryType = 0;
		uint32_t fileType = 0;
		uint32_t queryFileType = 0;
		int index_type = 0;

		if (strcmp(argv[4], "intersection") == 0) queryType = 0;
		else if (strcmp(argv[4], "10NN") == 0) queryType = 1;
		else if (strcmp(argv[4], "selfjoin") == 0) queryType = 2;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		if (strcmp(argv[2], "IdX1Y1X2Y2") == 0) fileType = INPUT_TYPE_ID_MBR;
		else if (strcmp(argv[2], "X1Y1X2Y2") == 0) fileType = INPUT_TYPE_MBR;
		else
		{
			cerr << "Unknown file type." << endl;
			return -1;
		}

		if (strcmp(argv[argc - 5], "IdX1Y1X2Y2") == 0) queryFileType = INPUT_TYPE_ID_MBR;
		else if (strcmp(argv[argc - 5], "X1Y1X2Y2") == 0) queryFileType = INPUT_TYPE_MBR;
		else
		{
			cerr << "Unknown query file type." << endl;
			return -1;
		}

		if (strcmp(argv[argc - 4], "REG") == 0) index_type = SRTREE_TYPE_REG;
		else if (strcmp(argv[argc - 4], "COMP") == 0) index_type = SRTREE_TYPE_COMP;
		else if (strcmp(argv[argc - 4], "HYB") == 0) index_type = SRTREE_TYPE_HYB;
		else
		{
			cerr << "Unknown index type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open data file " << argv[1] << "." << endl;
			return -1;
		}

		cerr << "Creating STR R-Tree..." << endl;
		
		// Create a new storage manager
		IStorageManager* memorymanager = StorageManager::createNewMemoryStorageManager();

		//Create the data stream
		MyDataStream stream(argv[1], fileType);

		// Create a new, empty, RTree with dimensionality 2, minimum load 70%, using "memorymanager" as
		// the StorageManager and the RSTAR splitting policy.
		id_type indexIdentifier;
		ISpatialIndex* tree = RTree::createAndBulkLoadNewRTree(
			RTree::BLM_STR, stream, 
			*memorymanager, 0.7, atoi(argv[3]), atoi(argv[3]), 2, SpatialIndex::RTree::RV_RSTAR, indexIdentifier);
			
		cerr << "Done." << endl;
		
		//cerr << "Operations: " << i << endl;
		cerr << *tree;
		//cerr << "Index ID: " << indexIdentifier << endl;

		//bool ret = tree->isIndexValid();
		//if (ret == false) cerr << "ERROR: Structure is invalid!" << endl;
		//else cerr << "The stucture seems O.K." << endl;
		
		cout << "Transforming to our tree..." << endl;
		RTree::SRTree* tSRTree = new RTree::SRTree((RTree::RTree*)tree, atoi(argv[argc-2]), index_type, atoi(argv[argc-3]));
		cout << "Done." << endl;
		delete tree;
		int tB,cB,iB,sB;
		tSRTree->memory_usage(tB,cB,iB,sB);
		cout << tB << ";" << cB << ";" << iB << ";" << sB << endl;

		long* resultSRTree = new long[tSRTree->getNumObj()];
		int resultSRTreeSize;
		
		FILE* outputFile = fopen(argv[5], "a");
		uint k;
		for(k = 6; k < argc - 5; k++) {
			//read input queries
			uint number_of_input_queries;
			INPUT_MBR *input_queries;
			if(queryFileType == INPUT_TYPE_ID_MBR) {
				readIdX1Y1X2Y2(argv[k], &input_queries, &number_of_input_queries);
			} else if(queryFileType == INPUT_TYPE_MBR) {
				readX1Y1X2Y2(argv[k], &input_queries, &number_of_input_queries);
			}
			
			//execute queries
			double time = 0;
		
			int i;
			double plow[2], phigh[2];
			for(i = 0; i < number_of_input_queries; i++) {
				plow[0] = input_queries[i].mbr.point1.lon; plow[1] = input_queries[i].mbr.point1.lat;
				phigh[0] = input_queries[i].mbr.point2.lon; phigh[1] = input_queries[i].mbr.point2.lat;
				Region r = Region(plow, phigh, 2);
				
				tTiempo mitiempo1;
				tTiempo mitiempo2;
				
				register uint j;

				getTiempo(&mitiempo1);

				for(j = 0; j < 5; j++) {
				  
				  resultSRTreeSize = 0;
				  tSRTree->range_query(plow[0], plow[1], phigh[0], phigh[1], resultSRTree, &resultSRTreeSize);

				}
				
				getTiempo(&mitiempo2);
				time += ((mitiempo2.user - mitiempo1.user) / 5);
	
				if(strcmp(argv[argc - 1], "test") == 0) {
					cout << "Query " << input_queries[i].id << ": ";
					//vis.printResult();
					cout << resultSRTreeSize << " elements: ";
					vector<long> myvector(resultSRTree, resultSRTree+resultSRTreeSize);
					vector<long>::iterator it;
					sort (myvector.begin(), myvector.end());
					for (it=myvector.begin(); it!=myvector.end(); ++it)
					  cout << *it << "; ";
					cout << endl;
				}
			}
			fprintf(outputFile, "%lf;", time / number_of_input_queries);
			free(input_queries);
		}

		fprintf(outputFile, "%d;%d;", 20*atoi(argv[3])*stream.size()/(atoi(argv[3])-1), tSRTree->memory_usage());//space: sizeof(entry)*m*n/m-1
		fprintf(outputFile, "\n");
		fclose(outputFile);
		
		delete [] resultSRTree;
		delete tSRTree;
		delete memorymanager;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail trying to write the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
